/* bytes: count bytes. */
/* Original date: 2024-06-30 */
/* Version: 2.9 */
/* Date: 2024-10-31 */

#include <ctype.h>
#include <errno.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>


/* Read a line from a file. */
static int
get_line (char **linep, size_t *sizep, FILE * const fp)
{
  char *line = *linep;
  size_t size = *sizep;
  int c;
  bool skip = false;
  size_t k = 0;
  if (line == NULL)
    {
      size = 1;
      line = malloc (size);
      if (line == NULL)
	{
	  return EOF;
	}
      else
	{
	  *linep = line;
	  *sizep = size;
	}
    }
  while ((c = fgetc (fp)) != EOF && c != '\n')
    {
      if (skip == false)
	{
	  if (c == '#')
	    {
	      skip = true;
	    }
	  else
	    {
	      if (k == size - 1)
		{
		  size = size * 1.1 + 1;
		  line = realloc (*linep, size);
		  if (line == NULL)
		    {
		      *(*linep + k) = '\0';
		      return EOF;
		    }
		  else
		    {
		      *linep = line;
		      *sizep = size;
		    }
		}
	      line[k++] = c;
	    }
	}
    }
  line[k] = '\0';
  return c;
}

/* Read a string from a file. */
static int
get_string (char **strp, size_t *sizep, FILE * const fp)
{
  char *str = *strp;
  size_t size = *sizep;
  int c;
  size_t k = 0;
  if (str == NULL)
    {
      size = 1;
      str = malloc (size);
      if (str == NULL)
	{
	  return EOF;
	}
      else
	{
	  *strp = str;
	  *sizep = size;
	}
    }
  while ((c = fgetc (fp)) != EOF && c != '\0')
    {
      if (k == size - 1)
	{
	  size = size * 1.1 + 1;
	  str = realloc (*strp, size);
	  if (str == NULL)
	    {
	      *(*strp + k) = '\0';
	      return EOF;
	    }
	  else
	    {
	      *strp = str;
	      *sizep = size;
	    }
	}
      str[k++] = c;
    }
  str[k] = '\0';
  return c;
}

/* Convert an xdigit to a value. */
static unsigned char
hex (const char c)
{
  if (c == '0')
    {
      return 0;
    }
  else if (c == '1')
    {
      return 1;
    }
  else if (c == '2')
    {
      return 2;
    }
  else if (c == '3')
    {
      return 3;
    }
  else if (c == '4')
    {
      return 4;
    }
  else if (c == '5')
    {
      return 5;
    }
  else if (c == '6')
    {
      return 6;
    }
  else if (c == '7')
    {
      return 7;
    }
  else if (c == '8')
    {
      return 8;
    }
  else if (c == '9')
    {
      return 9;
    }
  else if (c == 'A' || c == 'a')
    {
      return 10;
    }
  else if (c == 'B' || c == 'b')
    {
      return 11;
    }
  else if (c == 'C' || c == 'c')
    {
      return 12;
    }
  else if (c == 'D' || c == 'd')
    {
      return 13;
    }
  else if (c == 'E' || c == 'e')
    {
      return 14;
    }
  else
    {
      return 15;
    }
}

/* Process a format string. */
static void
process_format (const unsigned char *restrict format_str,
		const size_t format_len, const char *const restrict file_name,
		const int i, const unsigned long long int count,
		const unsigned long long int total, const double proportion)
{
  if (format_str != NULL)
    {
      const unsigned char *f = format_str;
      bool percent = false;
      while (f < format_str + format_len)
	{
	  const unsigned char c = *f++;
	  if (percent == true)
	    {
	      percent = false;
	      if (c == '%')
		{
		  putchar ('%');
		}
	      else if (c == 'A')
		{
		  if (i == '\'')
		    {
		      putchar ('\\');
		    }
		}
	      else if (c == 'B')
		{
		  if (i == '\\')
		    {
		      putchar ('\\');
		    }
		}
	      else if (c == 'b')
		{
		  unsigned char j = 128;
		  do
		    {
		      if ((i & j) != 0)
			{
			  putchar ('1');
			}
		      else
			{
			  putchar ('0');
			}
		      j /= 2;
		    }
		  while (j != 0);
		}
	      else if (c == 'c')
		{
		  putchar (i);
		}
	      else if (c == 'd')
		{
		  printf ("%d", i);
		}
	      else if (c == 'E' || c == 'e')
		{
		  if (isprint (i) != 0)
		    {
		      putchar (i);
		    }
		  else if (i == '\a')
		    {
		      printf ("\\a");
		    }
		  else if (i == '\b')
		    {
		      printf ("\\b");
		    }
		  else if (i == '\t')
		    {
		      printf ("\\t");
		    }
		  else if (i == '\n')
		    {
		      printf ("\\n");
		    }
		  else if (i == '\v')
		    {
		      printf ("\\v");
		    }
		  else if (i == '\f')
		    {
		      printf ("\\f");
		    }
		  else if (i == '\r')
		    {
		      printf ("\\r");
		    }
		  else if (c == 'E')
		    {
		      printf ("\\x%02X", i);
		    }
		  else
		    {
		      printf ("\\%o", i);
		    }
		}
	      else if (c == 'f')
		{
		  printf ("%s", file_name);
		}
	      else if (c == 'n')
		{
		  printf ("%llu", count);
		}
	      else if (c == 'O')
		{
		  printf ("%03o", i);
		}
	      else if (c == 'o')
		{
		  printf ("%#o", i);
		}
	      else if (c == 'P')
		{
		  printf ("%0.2f", 100 * proportion);
		}
	      else if (c == 'p')
		{
		  printf ("%f", proportion);
		}
	      else if (c == 'Q')
		{
		  if (i == '"')
		    {
		      putchar ('\\');
		    }
		}
	      else if (c == 'S')
		{
		  if (i == ' ')
		    {
		      putchar ('\\');
		    }
		}
	      else if (c == 't')
		{
		  printf ("%llu", total);
		}
	      else if (c == 'X')
		{
		  printf ("%02X", i);
		}
	      else if (c == 'x')
		{
		  printf ("%02x", i);
		}
	      else
		{
		  putchar ('%');
		  putchar (c);
		}
	    }
	  else if (c == '%')
	    {
	      percent = true;
	    }
	  else
	    {
	      putchar (c);
	    }
	}
      if (percent == true)
	{
	  percent = false;
	  putchar ('%');
	}
    }
}

/* Output results. */
static void
output (const char *const restrict file_name,
	const bool calculate_total,
	const bool calculate_proportion,
	const bool sort,
	unsigned long long int min_count,
	unsigned long long int max_count,
	const double min_proportion,
	const double max_proportion,
	const unsigned char *const restrict begin_format_str,
	const size_t begin_format_len,
	const unsigned char *const restrict item_format_str,
	const size_t item_format_len,
	const unsigned char *const restrict end_format_str,
	const size_t end_format_len, const unsigned long long int array[256])
{
  if (begin_format_str != NULL || item_format_str != NULL
      || end_format_str != NULL)
    {
      unsigned long long int total = 0;
      if (calculate_total == true || calculate_proportion == true)
	{
	  unsigned short int i;
	  for (i = 0; i < 256; i++)
	    {
	      total += array[i];
	    }
	}
      process_format (begin_format_str, begin_format_len, file_name, 0, 0,
		      total, 0);
      if (item_format_str != NULL && *item_format_str != '\0')
	{
	  if (calculate_total == true || calculate_proportion == true)
	    {
	      if (min_proportion > 0)
		{
		  const unsigned long long int count =
		    min_proportion * total + 0.5;
		  if (count > min_count)
		    {
		      min_count = count;
		    }
		}
	      if (max_proportion < 1)
		{
		  const unsigned long long int count = max_proportion * total;
		  if (max_count == 0 || count < max_count)
		    {
		      max_count = count;
		    }
		}
	    }
	  if (calculate_total == false || total >= min_count)
	    {
	      unsigned char item[256];
	      unsigned short int count = 0;
	      unsigned short int i;
	      for (i = 0; i < 256; i++)
		{
		  if (array[i] >= min_count
		      && (max_count == 0 || array[i] <= max_count))
		    {
		      item[count++] = i;
		    }
		}
	      if (count != 0)
		{
		  if (sort == true)
		    {
		      unsigned char g;
		      for (g = count / 2; g > 0; g /= 2)
			{
			  const unsigned char h = count - g;
			  unsigned char i;
			  for (i = 0; i < g; i++)
			    {
			      unsigned char j = i;
			      do
				{
				  const unsigned char k = j + g;
				  if (array[item[j]] < array[item[k]]
				      || (array[item[j]] == array[item[k]]
					  && item[j] > item[k]))
				    {
				      const unsigned char t = item[j];
				      item[j] = item[k];
				      item[k] = t;
				      if (j != i)
					{
					  j -= g;
					}
				      else
					{
					  j = k;
					}
				    }
				  else
				    {
				      j = k;
				    }
				}
			      while (j < h);
			    }
			}
		    }
		}
	      if (calculate_proportion == true)
		{
		  for (i = 0; i < count; i++)
		    {
		      const unsigned char j = item[i];
		      process_format (item_format_str, item_format_len,
				      file_name, j, array[j], total,
				      1.0 * array[j] / total);
		    }
		}
	      else
		{
		  for (i = 0; i < count; i++)
		    {
		      const unsigned char j = item[i];
		      process_format (item_format_str, item_format_len,
				      file_name, j, array[j], total, 0);
		    }
		}
	    }
	}
      process_format (end_format_str, end_format_len, file_name, 0, 0, total,
		      0);
    }
}

/* Count the frequncy of each byte in a file. */
static bool
read_file (const char *const restrict program_name,
	   const char *const restrict file_name,
	   const bool file_total,
	   const bool calculate_proportion,
	   const bool sort,
	   const unsigned long long int min_count,
	   const unsigned long long int max_count,
	   const double min_proportion,
	   const double max_proportion,
	   const unsigned char *const restrict begin_format_str,
	   const size_t begin_format_len,
	   const unsigned char *const restrict item_format_str,
	   const size_t item_format_len,
	   const unsigned char *const restrict end_format_str,
	   const size_t end_format_len,
	   const bool combine, unsigned long long int combined[256])
{
  bool return_value = true;
  FILE *fp = stdin;
  if (strcmp (file_name, "-") != 0)
    {
      fp = fopen (file_name, "r");
      if (fp == NULL)
	{
	  fprintf (stderr, "%s: ", program_name);
	  perror (file_name);
	  return_value = false;
	}
    }
  if (fp != NULL)
    {
      unsigned long long int current[256] = { 0 };
      int c;
      while ((c = fgetc (fp)) != EOF)
	{
	  current[c]++;
	}
      if (feof (fp) == 0)
	{
	  fprintf (stderr, "%s: ", program_name);
	  perror (file_name);
	  return_value = false;
	}
      else
	{
	  if (combine == true)
	    {
	      for (c = 0; c < 256; c++)
		{
		  combined[c] += current[c];
		}
	    }
	  output (file_name, file_total, calculate_proportion, sort,
		  min_count, max_count, min_proportion, max_proportion,
		  begin_format_str, begin_format_len, item_format_str,
		  item_format_len, end_format_str, end_format_len, current);
	}
      if (fp != stdin)
	{
	  if (0 != fclose (fp))
	    {
	      fprintf (stderr, "%s: ", program_name);
	      perror (file_name);
	      return_value = false;
	    }
	}
    }
  return return_value;
}

/* Read a file that contains a list of file names. */
static bool
read_file_list (const char *const restrict program_name,
		const char *const restrict file_name,
		const bool line_by_line,
		const bool file_total,
		const bool file_proportion,
		const bool sort,
		const unsigned long long int min_count,
		const unsigned long long int max_count,
		const double min_proportion,
		const double max_proportion,
		const unsigned char *const restrict begin_format_str,
		const size_t begin_format_len,
		const unsigned char *const restrict item_format_str,
		const size_t item_format_len,
		const unsigned char *const restrict end_format_str,
		const size_t end_format_len,
		const bool combine,
		unsigned long long int combined[256],
		char **strp, size_t *sizep)
{
  bool return_value = true;
  FILE *fp = stdin;
  if (strcmp (file_name, "-") != 0)
    {
      fp = fopen (file_name, "r");
      if (fp == NULL)
	{
	  fprintf (stderr, "%s: ", program_name);
	  perror (file_name);
	  return_value = false;
	}
    }
  if (fp != NULL)
    {
      size_t file_num = 1;
      if (line_by_line == true)
	{
	  while (get_line (strp, sizep, fp) == '\n')
	    {
	      if (**strp != '\0')
		{
		  if (read_file
		      (program_name, *strp, file_total, file_proportion, sort,
		       min_count, max_count, min_proportion, max_proportion,
		       begin_format_str, begin_format_len, item_format_str,
		       item_format_len, end_format_str, end_format_len,
		       combine, combined) == false)
		    {
		      return_value = false;
		    }
		}
	      file_num++;
	    }
	}
      else
	{
	  while (get_string (strp, sizep, fp) == '\0')
	    {
	      if (**strp != '\0')
		{
		  if (read_file
		      (program_name, *strp, file_total, file_proportion, sort,
		       min_count, max_count, min_proportion, max_proportion,
		       begin_format_str, begin_format_len, item_format_str,
		       item_format_len, end_format_str, end_format_len,
		       combine, combined) == false)
		    {
		      return_value = false;
		    }
		}
	      file_num++;
	    }
	}
      if (feof (fp) == 0)
	{
	  fprintf (stderr, "%s: %s:%zu: ", program_name, file_name, file_num);
	  perror (NULL);
	  return_value = false;
	}
      if (fp != stdin)
	{
	  if (0 != fclose (fp))
	    {
	      fprintf (stderr, "%s: ", program_name);
	      perror (file_name);
	      return_value = false;
	    }
	}
    }
  return return_value;
}

/* Copy a string to a buffer and replace backslash escape sequences. */
static bool
copy_string (const char *s, unsigned char **bp, size_t *sizep)
{
  const size_t len = strlen (s);
  if (len == 0)
    {
      if (*bp != NULL)
	{
	  free (*bp);
	  *bp = NULL;
	}
      *sizep = 0;
    }
  else
    {
      unsigned char *buf = realloc (*bp, len);
      if (buf == NULL)
	{
	  return false;
	}
      else
	{
	  size_t size = 0;
	  bool control = false;
	  *bp = buf;
	  do
	    {
	      unsigned char c = *s++;
	      if (c == '\\')
		{
		  if (*s == '\\')
		    {
		      s++;
		    }
		  else if (*s >= '0' && *s <= '7')
		    {
		      unsigned char i = 0;
		      c = *s++ - '0';
		      while (i < 2 && *s >= '0' && *s <= '7')
			{
			  i++;
			  c *= 8;
			  c += *s++ - '0';
			}
		    }
		  else if ((*s == 'U' || *s == 'u') && isxdigit (s[1]) != 0)
		    {
		      const unsigned char max = (*s++ == 'U') ? 7 : 3;
		      unsigned long int u = hex (*s++);
		      unsigned char i = 0;
		      while (i < max && isxdigit (*s) != 0)
			{
			  i++;
			  u *= 16;
			  u += hex (*s++);
			}
		      if (u < 128)
			{
			  c = u;
			}
		      else
			{
			  char t[6] = { 0 };
			  unsigned char f = '\x80';
			  i = 0;
			  c = '\x40';
			  do
			    {
			      f |= c;
			      c /= 2;
			      t[i++] = ((u & '\x3F') | '\x80');
			      u >>= 6;
			    }
			  while (u >= c);
			  c--;
			  buf[size] = ((u & c) | f);
			  if (control == true)
			    {
			      control = false;
			      buf[size] ^= '\x40';
			    }
			  size++;
			  do
			    {
			      buf[size++] = t[--i];
			    }
			  while (i > 0);
			  continue;
			}
		    }
		  else if (*s == 'a')
		    {
		      s++;
		      c = '\a';
		    }
		  else if (*s == 'b')
		    {
		      s++;
		      c = '\b';
		    }
		  else if (*s == 'c' && (s[1] != '\0' || control == true))
		    {
		      s++;
		      control = !control;
		      continue;
		    }
		  else if (*s == 'd' && isdigit (s[1]) != 0)
		    {
		      unsigned char i = 0;
		      s++;
		      c = *s++ - '0';
		      while (i < 2 && isdigit (*s) != 0)
			{
			  i++;
			  c *= 10;
			  c += *s++ - '0';
			}
		    }
		  else if (*s == 'f')
		    {
		      s++;
		      c = '\f';
		    }
		  else if (*s == 'n')
		    {
		      s++;
		      c = '\n';
		    }
		  else if (*s == 'o' && s[1] >= '0' && s[1] <= '7')
		    {
		      unsigned char i = 0;
		      s++;
		      c = *s++ - '0';
		      while (i < 2 && *s >= '0' && *s <= '7')
			{
			  i++;
			  c *= 8;
			  c += *s++ - '0';
			}
		    }
		  else if (*s == 'r')
		    {
		      s++;
		      c = '\r';
		    }
		  else if (*s == 't')
		    {
		      s++;
		      c = '\t';
		    }
		  else if (*s == 'v')
		    {
		      s++;
		      c = '\v';
		    }
		  else if (*s == 'x' && isxdigit (s[1]) != 0)
		    {
		      s++;
		      c = hex (*s++);
		      if (isxdigit (*s) != 0)
			{
			  c *= 16;
			  c += hex (*s++);
			}
		    }
		}
	      if (control == true)
		{
		  control = false;
		  if (islower (c) != 0)
		    {
		      c = toupper (c);
		    }
		  c ^= '\x40';
		}
	      buf[size++] = c;
	    }
	  while (*s != '\0');
	  *sizep = size;
	  if (size != len)
	    {
	      buf = realloc (*bp, size);
	      if (buf == NULL)
		{
		  return false;
		}
	      else
		{
		  *bp = buf;
		}
	    }
	}
    }
  return true;
}

/* Check whether the format strings contain a character in the string. */
static bool
format_uses (const unsigned char *restrict format_str,
	     const size_t format_len, const char *const restrict s)
{
  if (format_str != NULL)
    {
      const unsigned char *f = format_str;
      bool percent = false;
      while (f < format_str + format_len)
	{
	  const unsigned char c = *f++;
	  if (percent == true)
	    {
	      percent = false;
	      if (strchr (s, c) != NULL)
		{
		  return true;
		}
	    }
	  else if (c == '%')
	    {
	      percent = true;
	    }
	}
    }
  return false;
}

int
main (int argc, char *argv[])
{
  const char *const program_name = argv[0];
  int return_value = EXIT_SUCCESS;
  unsigned long long int combined[256] = { 0 };
  enum
  { sort, last_boolean_opt };
  enum
  { max_count, min_count, last_unsigned_opt };
  enum
  { max_proportion, min_proportion, last_numeric_opt };
  enum
  { begin_file_format, item_file_format, end_file_format,
      begin_combined_format, item_combined_format, end_combined_format,
      last_string_opt };
  bool boolean_opt[last_boolean_opt] = { false };
  unsigned long long int unsigned_opt[last_unsigned_opt] = { 0 };
  double numeric_opt[last_numeric_opt] = { 0.0 };
  unsigned char *string_opt[last_string_opt] = { NULL };
  size_t string_opt_len[last_string_opt] = { 0 };
  unsigned_opt[min_count] = 1;
  numeric_opt[max_proportion] = 1;
  bool custom_format = false;
  bool file_total = false;
  bool combined_total = false;
  bool file_proportion = false;
  bool combined_proportion = false;
  bool combine = false;
  bool end = false;
  unsigned long int file_count = 0;
  if (argc > 1)
    {
      enum opt_type
      { info_opt_type, boolean_opt_type, unsigned_opt_type, numeric_opt_type,
	  string_opt_type, list_opt_type };
      enum
      { help, version, last_info_opt };
      enum
      { null_terminated, newline_terminated, last_list_opt };
      char *str = NULL;
      size_t size = 0;
      struct opt
      {
	const char *const short_opt;
	const char *const long_opt;
	const char *const help;
	unsigned char len;
	const int type;
	const int item;
      };
      struct opt opt_array[] = {
	{"L FILE", "null-list=FILE",
	 "Read the list of files in FILE. A null character follows each file name.",
	 0, list_opt_type, null_terminated},
	{"l FILE", "list=FILE",
	 "Read the list of files in FILE. Each line of the file contains a file name.",
	 0, list_opt_type, newline_terminated},
	{"N NUMBER", "maximum-count=NUMBER",
	 "Set a maximum count. Output nothing for bytes that occur more than this.",
	 0, unsigned_opt_type, max_count},
	{"n NUMBER", "minimum-count=NUMBER",
	 "Set a minimum count. Output nothing for bytes that occur less than this. The default minimum count is 1.",
	 0, unsigned_opt_type, min_count},
	{"P NUMBER", "maximum-proportion=NUMBER",
	 "Set a maximum proportion. Output nothing for bytes that occur more than this.",
	 0, numeric_opt_type, max_proportion},
	{"p NUMBER", "minimum-proportion=NUMBER",
	 "Set a minimum proportion. Output nothing for bytes that occur less than this.",
	 0, numeric_opt_type, min_proportion},
	{"B STRING", "begin-combined-format=STRING",
	 "Set the output format string to use before the combined output for all files.",
	 0, string_opt_type, begin_combined_format},
	{"F STRING", "item-combined-format=STRING",
	 "Set the output format string to use for each byte in the combined output for all files.",
	 0, string_opt_type, item_combined_format},
	{"E STRING", "end-combined-format=STRING",
	 "Set the output format string to use after the combined output for all files.",
	 0, string_opt_type, end_combined_format},
	{"b STRING", "begin-file-format=STRING",
	 "Set the output format string to use before the output for each file.",
	 0, string_opt_type, begin_file_format},
	{"f STRING", "item-file-format=STRING",
	 "Set the output format string to use for each byte in the output for each file.",
	 0, string_opt_type, item_file_format},
	{"e STRING", "end-file-format=STRING",
	 "Set the output format string to use after the output for each file.",
	 0, string_opt_type, end_file_format},
	{"s", "sort", "Sort bytes by count (highest first).", 0,
	 boolean_opt_type, sort},
	{"h", "help", "Output this help and exit.", 0, info_opt_type, help},
	{"V", "version", "Output version information and exit.", 0,
	 info_opt_type, version},
      };
      const size_t opt_count = sizeof (opt_array) / sizeof (opt_array[0]);
      size_t j;
      for (j = 0; j < opt_count; j++)
	{
	  const int type = opt_array[j].type;
	  const int item = opt_array[j].item;
	  if (type == info_opt_type && item < last_info_opt)
	    {
	    }
	  else if (type == boolean_opt_type && item < last_boolean_opt)
	    {
	    }
	  else if (type == unsigned_opt_type && item < last_unsigned_opt)
	    {
	    }
	  else if (type == numeric_opt_type && item < last_numeric_opt)
	    {
	    }
	  else if (type == string_opt_type && item < last_string_opt)
	    {
	    }
	  else if (type == list_opt_type && item < last_list_opt)
	    {
	    }
	  else
	    {
	      fprintf (stderr,
		       "%s: internal error: invalid option: %zu (type %d, item %d)\n",
		       program_name, j, type, item);
	      return_value = EXIT_FAILURE;
	      end = true;
	    }
	  if (opt_array[j].long_opt != NULL)
	    {
	      const char *const s = strchr (opt_array[j].long_opt, '=');
	      if (s != NULL)
		{
		  opt_array[j].len = s - opt_array[j].long_opt;
		}
	      else
		{
		  opt_array[j].len = strlen (opt_array[j].long_opt);
		}
	    }
	}
      unsigned int pass;
      for (pass = 0; pass < 2 && end == false; pass++)
	{
	  bool opt_end = false;
	  int i;
	  for (i = 1; i < argc && end == false; i++)
	    {
	      if (opt_end == false && argv[i][0] == '-' && argv[i][1] != '\0')
		{
		  if (argv[i][1] == '-' && argv[i][2] == '\0')
		    {
		      opt_end = true;
		    }
		  else
		    {
		      size_t opt_match_count = 0;
		      const char *s = argv[i] + 1;
		      const char *v = NULL;
		      bool more = false;
		      do
			{
			  size_t opt_match = 0;
			  if (more == false && *s == '-')
			    {
			      s++;
			      size_t j;
			      size_t len = strlen (s);
			      v = strchr (s, '=');
			      if (v != NULL)
				{
				  len = v - s;
				  v++;
				}
			      for (j = 0; j < opt_count; j++)
				{
				  if (opt_array[j].long_opt != NULL)
				    {
				      if (v == NULL
					  || opt_array[j].
					  long_opt[opt_array[j].len] == '=')
					{
					  if (len == opt_array[j].len
					      && strncmp (s,
							  opt_array[j].
							  long_opt, len) == 0)
					    {
					      opt_match_count = 1;
					      opt_match = j;
					      break;
					    }
					  else if (len < opt_array[j].len
						   && strncmp (s,
							       opt_array[j].
							       long_opt,
							       len) == 0)
					    {
					      if (opt_match_count == 0)
						{
						  opt_match_count = 1;
						  opt_match = j;
						}
					      else
						if ((opt_array[j].type !=
						     opt_array[opt_match].
						     type)
						    || (opt_array[j].item !=
							opt_array[opt_match].
							item))
						{
						  opt_match_count++;
						}
					    }
					}
				    }
				}
			      if (opt_match_count == 0)
				{
				  fprintf (stderr,
					   "%s: unrecognized option '%s'\n",
					   program_name, argv[i]);
				  return_value = EXIT_FAILURE;
				  end = true;
				}
			      else if (opt_match_count == 1)
				{
				  if (v == NULL)
				    {
				      if (opt_array[opt_match].
					  long_opt[opt_array[opt_match].
						   len] == '=')
					{
					  if (i + 1 < argc)
					    {
					      v = argv[++i];
					    }
					  else
					    {
					      fprintf (stderr,
						       "%s: option --%.*s must have an argument\n",
						       program_name,
						       opt_array[opt_match].
						       len,
						       opt_array[opt_match].
						       long_opt);
					      return_value = EXIT_FAILURE;
					      end = true;
					    }
					}
				    }
				}
			      else
				{
				  size_t j;
				  fprintf (stderr,
					   "%s: option '%s' is ambiguous: '--%.*s'",
					   program_name, argv[i],
					   opt_array[opt_match].len,
					   opt_array[opt_match].long_opt);
				  for (j = opt_match + 1; j < opt_count; j++)
				    {
				      if (opt_array[j].long_opt != NULL)
					{
					  if (v == NULL
					      || opt_array[j].
					      long_opt[opt_array[j].len] ==
					      '=')
					    {
					      if (len < opt_array[j].len
						  && strncmp (s,
							      opt_array[j].
							      long_opt,
							      len) == 0)
						{
						  fprintf (stderr,
							   ", '--%.*s'",
							   opt_array[j].len,
							   opt_array[j].
							   long_opt);
						}
					    }
					}
				    }
				  fputc ('\n', stderr);
				  return_value = EXIT_FAILURE;
				  end = true;
				}
			    }
			  if (end == false)
			    {
			      if (opt_match_count == 0)
				{
				  const unsigned c = *s++;
				  size_t j;
				  more = false;
				  for (j = 0; j < opt_count; j++)
				    {
				      if (opt_array[j].short_opt != NULL)
					{
					  if (c == opt_array[j].short_opt[0])
					    {
					      opt_match = j;
					      if (opt_array[opt_match].
						  short_opt[1] == ' ')
						{
						  if (*s != '\0')
						    {
						      v = s;
						      s = strchr (s, '\0');
						    }
						  else if (i + 1 < argc)
						    {
						      v = argv[++i];
						    }
						  else
						    {
						      fprintf (stderr,
							       "%s: option %c must have an argument\n",
							       program_name,
							       c);
						      return_value =
							EXIT_FAILURE;
						      end = true;
						    }
						}
					      else if (*s != '\0')
						{
						  more = true;
						}
					      break;
					    }
					}
				    }
				  if (j == opt_count)
				    {
				      if (isprint (c) != 0)
					{
					  fprintf (stderr,
						   "%s: invalid option '%c'\n",
						   program_name, c);
					}
				      else
					{
					  fprintf (stderr,
						   "%s: invalid option '\\x%02X'\n",
						   program_name, c);
					}
				      return_value = EXIT_FAILURE;
				      end = true;
				    }
				}
			      if (end == false)
				{
				  const int type = opt_array[opt_match].type;
				  const int item = opt_array[opt_match].item;
				  if (pass == 0)
				    {
				      if (type == info_opt_type)
					{
					  if (item == help)
					    {
					      size_t j;
					      printf
						("Usage: %s [OPTIONS] [FILES]\n",
						 program_name);
					      puts ("Count bytes.\n");
					      puts
						("With no file names or with the file name -, read standard input.\n");
					      for (j = 0; j < opt_count; j++)
						{
						  if (opt_array[j].
						      short_opt != NULL
						      && opt_array[j].
						      long_opt != NULL)
						    {
						      printf ("-%s, --%s\n",
							      opt_array[j].
							      short_opt,
							      opt_array[j].
							      long_opt);
						    }
						  else if (opt_array[j].
							   short_opt != NULL)
						    {
						      printf ("-%s\n",
							      opt_array[j].
							      short_opt);
						    }
						  else if (opt_array[j].
							   long_opt != NULL)
						    {
						      printf ("--%s\n",
							      opt_array[j].
							      long_opt);
						    }
						  if (opt_array[j].help !=
						      NULL)
						    {
						      puts (opt_array[j].
							    help);
						    }
						}
					      puts
						("\nOutput format strings can use %% for a percent sign.");
					      puts
						("Combined output format strings can use %t for the total byte count in all files.");
					      puts
						("File output format strings can use %f for the file name and %t for the total byte count in the file.");
					      puts
						("\nItem output format strings can also use:");
					      puts
						("%A for a backslash if the current byte is an apostrophe.");
					      puts
						("%B for a backslash if the current byte is a backslash.");
					      puts
						("%b for the binary value of the current byte.");
					      puts
						("%c for the current character.");
					      puts
						("%d for the decimal value of the current byte.");
					      puts
						("%E for the current character (if printable) or a hexadecimal escape sequence.");
					      puts
						("%e for the current character (if printable) or an octal escape sequence.");
					      puts
						("%n for the count of the current byte.");
					      puts
						("%O for the octal value of the current byte as 3 octal digits.");
					      puts
						("%o for the octal value of the current byte.");
					      puts
						("%P for the proportion of the current byte as a percentage.");
					      puts
						("%p for the proportion of the current byte.");
					      puts
						("%Q for a backslash if the current byte is a quotation mark.");
					      puts
						("%S for a backslash if the current byte is a space.");
					      puts
						("%X for the hexadecimal value of the current byte.");
					      puts
						("%x for the lowercase hexadecimal value of the current byte.");
					      puts
						("\nOutput format strings can use backslash escape sequences:");
					      puts ("\\\\ for backslash.");
					      puts
						("\\NNN for the byte with octal value NNN (1 to 3 digits).");
					      puts ("\\a for alert.");
					      puts ("\\b for backspace.");
					      puts
						("\\cC for control-C (convert a lowercase ASCII character to uppercase and flip bit 6).");
					      puts
						("\\dNNN for the byte with decimal value NNN (1 to 3 digits).");
					      puts ("\\f for form feed.");
					      puts ("\\n for newline.");
					      puts
						("\\oNNN for the byte with octal value NNN (1 to 3 digits).");
					      puts
						("\\r for carriage return.");
					      puts
						("\\t for horizontal tab.");
					      puts
						("\\UHHHHHHHH for the UTF-8 encoding of the Unicode character with hexadecimal value HHHHHHHH (1 to 8 digits).");
					      puts
						("\\uHHHH for the UTF-8 encoding of the Unicode character with hexadecimal value HHHH (1 to 4 digits).");
					      puts ("\\v for vertical tab.");
					      puts
						("\\xHH for the byte with hexadecimal value HH (1 to 2 digits).");
					      end = true;
					    }
					  else if (item == version)
					    {
					      puts ("bytes 2.9");
					    }
					  end = true;
					}
				      else if (type == list_opt_type)
					{
					  file_count += 2;
					}
				      else if (type == boolean_opt_type)
					{
					  boolean_opt[item] = true;
					}
				      else if (type == string_opt_type)
					{
					  custom_format = true;
					  if (copy_string
					      (v, &string_opt[item],
					       &string_opt_len[item]) ==
					      false)
					    {
					      fprintf (stderr, "%s: ",
						       program_name);
					      perror (v);
					      return_value = EXIT_FAILURE;
					      end = true;
					    }
					}
				      else if (type == unsigned_opt_type
					       || type == numeric_opt_type)
					{
					  if (*v == '\0')
					    {
					      if (opt_match_count == 1)
						{
						  fprintf (stderr,
							   "%s: option --%.*s must not have a blank argument\n",
							   program_name,
							   opt_array
							   [opt_match].len,
							   opt_array
							   [opt_match].
							   long_opt);
						}
					      else
						{
						  fprintf (stderr,
							   "%s: option %c must not have a blank argument\n",
							   program_name,
							   opt_array
							   [opt_match].
							   short_opt[0]);
						}
					      return_value = EXIT_FAILURE;
					      end = true;
					    }
					  else if (type == unsigned_opt_type)
					    {
					      char *t = NULL;
					      errno = 0;
					      const unsigned long long int
						value = strtoull (v, &t, 0);
					      if (*t != '\0')
						{
						  fprintf (stderr,
							   "%s: invalid positive integer: \"%s\"\n",
							   program_name, v);
						  return_value = EXIT_FAILURE;
						  end = true;
						}
					      else if (errno != 0)
						{
						  fprintf (stderr, "%s: ",
							   program_name);
						  perror (v);
						  return_value = EXIT_FAILURE;
						  end = true;
						}
					      else
						{
						  unsigned_opt[item] = value;
						}
					    }
					  else if (type == numeric_opt_type)
					    {
					      char *t = NULL;
					      errno = 0;
					      double value = strtod (v, &t);
					      if (*t == '%' && t[1] == '\0')
						{
						  value /= 100;
						  t++;
						}
					      if (*t != '\0')
						{
						  fprintf (stderr,
							   "%s: invalid floating point number: \"%s\"\n",
							   program_name, v);
						  return_value = EXIT_FAILURE;
						  end = true;
						}
					      else if (errno != 0)
						{
						  fprintf (stderr, "%s: ",
							   program_name);
						  perror (v);
						  return_value = EXIT_FAILURE;
						  end = true;
						}
					      else if (value < 0)
						{
						  fprintf (stderr,
							   "%s: invalid value: proportion is less than 0: \"%s\"\n",
							   program_name, v);
						  return_value = EXIT_FAILURE;
						  end = true;
						}
					      else if (value > 1)
						{
						  fprintf (stderr,
							   "%s: invalid value: proportion is greater than 1: \"%s\"\n",
							   program_name, v);
						  return_value = EXIT_FAILURE;
						  end = true;
						}
					      else
						{
						  numeric_opt[item] = value;
						}
					    }
					}
				    }
				  else
				    {
				      if (type == list_opt_type)
					{
					  if (read_file_list
					      (program_name, v,
					       (item == newline_terminated),
					       file_total, file_proportion,
					       boolean_opt[sort],
					       unsigned_opt[min_count],
					       unsigned_opt[max_count],
					       numeric_opt[min_proportion],
					       numeric_opt[max_proportion],
					       string_opt[begin_file_format],
					       string_opt_len
					       [begin_file_format],
					       string_opt[item_file_format],
					       string_opt_len
					       [item_file_format],
					       string_opt[end_file_format],
					       string_opt_len
					       [end_file_format], combine,
					       combined, &str,
					       &size) == false)
					    {
					      return_value = EXIT_FAILURE;
					    }
					}
				    }
				}
			    }
			}
		      while (more == true && end == false);
		    }
		}
	      else if (pass == 0)
		{
		  file_count++;
		}
	      else
		{
		  if (read_file
		      (program_name, argv[i], file_total, file_proportion,
		       boolean_opt[sort], unsigned_opt[min_count],
		       unsigned_opt[max_count], numeric_opt[min_proportion],
		       numeric_opt[max_proportion],
		       string_opt[begin_file_format],
		       string_opt_len[begin_file_format],
		       string_opt[item_file_format],
		       string_opt_len[item_file_format],
		       string_opt[end_file_format],
		       string_opt_len[end_file_format], combine,
		       combined) == false)
		    {
		      return_value = EXIT_FAILURE;
		    }
		}
	    }
	  if (pass == 0 && end == false)
	    {
	      if (custom_format == false)
		{
		  file_total = true;
		  if (file_count <= 1)
		    {
		      if (copy_string
			  ("%d '%A%B%E' %n\n", &string_opt[item_file_format],
			   &string_opt_len[item_file_format]) == false
			  || copy_string ("%t\n",
					  &string_opt[end_file_format],
					  &string_opt_len[end_file_format]) ==
			  false)
			{
			  perror (program_name);
			  return_value = EXIT_FAILURE;
			  end = true;
			}
		    }
		  else
		    {
		      combined_total = true;
		      combine = true;
		      if (copy_string
			  ("%f: %d '%A%B%E' %n\n",
			   &string_opt[item_file_format],
			   &string_opt_len[item_file_format]) == false
			  || copy_string ("%f: %t\n",
					  &string_opt[end_file_format],
					  &string_opt_len[end_file_format]) ==
			  false
			  || copy_string ("%d '%A%B%E' %n\n",
					  &string_opt[item_combined_format],
					  &string_opt_len
					  [item_combined_format]) == false
			  || copy_string ("%t\n",
					  &string_opt[end_combined_format],
					  &string_opt_len
					  [end_combined_format]) == false)
			{
			  perror (program_name);
			  return_value = EXIT_FAILURE;
			  end = true;
			}
		    }
		}
	      else
		{
		  if (format_uses
		      (string_opt[item_file_format],
		       string_opt_len[item_file_format], "Pp") == true)
		    {
		      file_proportion = true;
		    }
		  else
		    if (format_uses
			(string_opt[begin_file_format],
			 string_opt_len[begin_file_format], "t") == true
			|| format_uses (string_opt[item_file_format],
					string_opt_len[item_file_format],
					"t") == true
			|| format_uses (string_opt[end_file_format],
					string_opt_len[end_file_format],
					"t") == true)
		    {
		      file_total = true;
		    }
		  else if (numeric_opt[min_proportion] > 0
			   || numeric_opt[max_proportion] < 1)
		    {
		      file_total = true;
		    }
		  if (format_uses
		      (string_opt[item_combined_format],
		       string_opt_len[item_combined_format], "Pp") == true)
		    {
		      combined_proportion = true;
		      combine = true;
		    }
		  else
		    if (format_uses
			(string_opt[begin_combined_format],
			 string_opt_len[begin_combined_format], "t") == true
			|| format_uses (string_opt[item_combined_format],
					string_opt_len[item_combined_format],
					"t") == true
			|| format_uses (string_opt[end_combined_format],
					string_opt_len[end_combined_format],
					"t") == true)
		    {
		      combined_total = true;
		      combine = true;
		    }
		  else if (string_opt[begin_combined_format] != NULL
			   || string_opt[item_combined_format] != NULL
			   || string_opt[end_combined_format] != NULL)
		    {
		      combine = true;
		      if (numeric_opt[min_proportion] > 0
			  || numeric_opt[max_proportion] < 1)
			{
			  combined_total = true;
			}
		    }
		}
	    }
	}
      if (str != NULL)
	{
	  free (str);
	  str = NULL;
	}
    }
  if (end == false)
    {
      if (file_count == 0)
	{
	  if (custom_format == false)
	    {
	      file_total = true;
	      if (copy_string
		  ("%d '%A%B%E' %n\n", &string_opt[item_file_format],
		   &string_opt_len[item_file_format]) == false
		  || copy_string ("%t\n", &string_opt[end_file_format],
				  &string_opt_len[end_file_format]) == false)
		{
		  perror (program_name);
		  return_value = EXIT_FAILURE;
		  end = true;
		}
	    }
	  if (end == false)
	    {
	      if (read_file
		  (program_name, "-", file_total, file_proportion,
		   boolean_opt[sort], unsigned_opt[min_count],
		   unsigned_opt[max_count], numeric_opt[min_proportion],
		   numeric_opt[max_proportion], string_opt[begin_file_format],
		   string_opt_len[begin_file_format],
		   string_opt[item_file_format],
		   string_opt_len[item_file_format],
		   string_opt[end_file_format],
		   string_opt_len[end_file_format], combine,
		   combined) == false)
		{
		  return_value = EXIT_FAILURE;
		}
	    }
	}
      if (end == false)
	{
	  output ("", combined_total, combined_proportion, boolean_opt[sort],
		  unsigned_opt[min_count], unsigned_opt[max_count],
		  numeric_opt[min_proportion], numeric_opt[max_proportion],
		  string_opt[begin_combined_format],
		  string_opt_len[begin_combined_format],
		  string_opt[item_combined_format],
		  string_opt_len[item_combined_format],
		  string_opt[end_combined_format],
		  string_opt_len[end_combined_format], combined);
	}
    }
  if (0 != fclose (stdout))
    {
      perror (program_name);
      return_value = EXIT_FAILURE;
    }
  return return_value;
}
